package com.ca.directory.jxplorer;

import com.ca.commons.cbutil.*;
import com.ca.directory.jxplorer.search.SearchGUI;
import com.ca.directory.jxplorer.tree.SmartTree;
import com.ca.directory.jxplorer.editor.PostalAddressEditor;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;



/**
*   Sets up an advanced options dialog box that is accessable through the Options drop down menu.
*   It has five tabbed panes: the first is for the look and feel, the second is for log levels,
*   the third is for the log method, the fourth is for LDAP limit & timeout and the last is for the URL handling.
*   When the user clicks the 'Apply' button all of the changes are saved in the property file and reflected in JX.
*   @author Trudi.
*/

public class AdvancedOptions extends JDialog
{
    protected CBButton            	btnApply, btnReset, btnCancel, btnHelp;
    protected JTextField         	ldapLimit, ldapTimeout;
    protected JComboBox             urlCombo, logLevelCombo, logMethodCombo;
    protected CBPanel            	display;
    protected JTabbedPane        	tabbedPane;  
    protected JRadioButton[]     	lookAndFeel;
    protected String[]           	lookAndFeelVal;
    protected final String[]        logLevelVal = new String[] {CBIntText.get("Errors Only"), CBIntText.get("Basic"), CBIntText.get("Tree Operations"), CBIntText.get("Extensive"), CBIntText.get("All"), CBIntText.get("All + BER Trace")};
    protected final String[]        logMethodVal = new String[] {CBIntText.get("None"), CBIntText.get("Console"), CBIntText.get("File"), CBIntText.get("Console & File")};
    // need to save these in english for saving in the config file...
    protected final String[]        logMethodValEnglish = new String[] {"None", "Console", "File", "Console & File"};
    protected MainMenu           	mainMenu;
    protected final JXplorer     	jx;
	String 							dirImage;

    // utility constants for look and feel stuff
    protected static final int WINDOWS = 0;
    protected static final int JAVA = 1;
    protected static final int MOTIF = 2;
    protected static final int MAC = 3;


   /**
    *    Sets up the panel, adds the tabbed pane and the buttons to it.
    *    @param jxplorer a JXplorer object to update changes with the log level, log method & LDAP values. 
    *    @param mainMenu a MainMenu object to update the gui when the L&F is changed.
    */
                           
    public AdvancedOptions(JXplorer jxplorer, MainMenu mainMenu) 
    { 
       
        super(jxplorer);
        setModal(true);
        
        this.mainMenu = mainMenu;
        jx = jxplorer;

		dirImage = jx.getProperty("dir.images");
		
        setTitle(CBIntText.get("JXplorer Advanced Options"));        
        
        display = new CBPanel();
                
        btnApply = new CBButton(CBIntText.get("Apply"), CBIntText.get("Click here to apply the changes"));
        btnApply.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                       apply();                   
           }});

        btnReset = new CBButton(CBIntText.get("Reset"), CBIntText.get("Click here to reset the options"));
        btnReset.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                       reset();
           }});
           
        btnCancel = new CBButton(CBIntText.get("Cancel"), CBIntText.get("Click here to exit Advanced Options"));
        btnCancel.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                       quit();
           }});
           
        btnHelp = new CBButton(CBIntText.get("Help"), CBIntText.get("Click here for help"));    //TE: creates a new help button with a listener that will open JX help at appropriate location.
        CBHelpSystem.useDefaultHelp(btnHelp, "config.advanced");

        //TE: better way to implement keystroke listening...
        display.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("ENTER"), "enter");
        display.getInputMap(JComponent.WHEN_IN_FOCUSED_WINDOW).put(KeyStroke.getKeyStroke("ESCAPE"), "escape");
        display.getActionMap().put("enter", new MyAction(CBAction.ENTER));
        display.getActionMap().put("escape", new MyAction(CBAction.ESCAPE));

        tabbedPane = new JTabbedPane();
		
		lookAndFeelTab();           //TE: sets up the L&F tab.
		ldapLevels();               //TE: sets up the ldap levels tab.
		logLevel();                 //TE: sets up the log level tab.
		logMethod();                //TE: sets up the log method tab.
        urlTab();                   //TE: sets up the other tab

        JPanel buttonPanel = new JPanel();

        buttonPanel.add(btnApply); 
        buttonPanel.add(btnReset);
        buttonPanel.add(btnCancel);
        buttonPanel.add(btnHelp);
        display.addln(tabbedPane);
        display.addln(buttonPanel);

        setSize(300,280);   // make 400,340 to have one row
    
        getContentPane().add(display); 
    }



    /**
     * Apparently it is better to use key bindings rather than adding a KeyListener...
     * "for reacting in a special way to particular keys, you usually should use key
     * bindings instead of a key listener".
     * This class lets the user set the key as an int.  If a key is pressed and it
     * matches the assigned int, a check is done for if it is an escape or enter key.
     * (27 or 10).  If escape, the quit method is called.  If enter, the apply
     * method is called.
     * Bug 4646.
     * @author Trudi.
     */
    private class MyAction extends CBAction
    {
        /**
         * Calls super constructor.
         * @param key
         */
        public MyAction(int key)
        {
            super(key);
        }

        /**
         * quit is called if the Esc key pressed,
         * apply is called if Enter key is pressed.
         * @param e never used.
         */
        public void actionPerformed(ActionEvent e)
        {
            if (getKey() == ESCAPE)
                quit();
            else if (getKey() == ENTER)
                apply();
        }
    }



   /**
    *    Sets up the L&F radio buttons.    
    **/
    
    
    public void lookAndFeelTab()
    {                        
        lookAndFeel = new JRadioButton[] {new JRadioButton(CBIntText.get("Windows Look and Feel")), new JRadioButton(CBIntText.get("Java Look and Feel")), new JRadioButton(CBIntText.get("Motif Look and Feel")), new JRadioButton(CBIntText.get("Mac Look and Feel"))};
            
        String [] toolTip = new String[] {CBIntText.get("Sets the look and feel to: Windows"), CBIntText.get("Sets the look and feel to: Java"), CBIntText.get("Sets the look and feel to: Motif"), CBIntText.get("Sets the look and feel to: Apple Mac/OSX")};
            
        ButtonGroup lookAndFeelButtonGroup = new ButtonGroup();
        CBPanel lookAndFeelPanel = new CBPanel();
        
        lookAndFeelPanel.addln(new JLabel(CBIntText.get("Select a New Look & Feel: ")));
        lookAndFeelPanel.addln(new JLabel(" "));    //TE: white space.
        
		if(jx.isWindows())
            addLookAndFeelOption(lookAndFeelButtonGroup, WINDOWS, lookAndFeelPanel, toolTip);
        else
            lookAndFeel[WINDOWS].setSelected(false);

        addLookAndFeelOption(lookAndFeelButtonGroup, JAVA, lookAndFeelPanel, toolTip);
        addLookAndFeelOption(lookAndFeelButtonGroup, MOTIF, lookAndFeelPanel, toolTip);

        if(jx.isMac())
            addLookAndFeelOption(lookAndFeelButtonGroup, MAC, lookAndFeelPanel, toolTip);
        else
            lookAndFeel[MAC].setSelected(false);

        getLookAndFeel();


        tabbedPane.addTab(CBIntText.get("Look & Feel"), new ImageIcon(dirImage + "look_feel.gif"),lookAndFeelPanel, CBIntText.get("Change the 'look and feel' of JXplorer, that is, adopt a similar appearance to another application."));
    }

    private void addLookAndFeelOption(ButtonGroup lookAndFeelButtonGroup, int i, CBPanel lookAndFeelPanel, String[] toolTip)
    {
        lookAndFeelButtonGroup.add(lookAndFeel[i]);
        lookAndFeelPanel.addln(lookAndFeel[i]);
        lookAndFeel[i].setToolTipText(toolTip[i]);
    }


    /**
    *    Gets the L&F from the property file and set the appropriate combo box item.
    */
    
    public void getLookAndFeel()
    {              
        lookAndFeelVal = new String[] {"com.sun.java.swing.plaf.windows.WindowsLookAndFeel", "javax.swing.plaf.metal.MetalLookAndFeel", "com.sun.java.swing.plaf.motif.MotifLookAndFeel", "com.sun.java.swing.plaf.mac.MacLookAndFeel"};
        
        for(int i=0; i<4; i++)
        {
            if (String.valueOf(lookAndFeelVal[i]).equalsIgnoreCase(JXplorer.getProperty("gui.lookandfeel")))
                lookAndFeel[i].setSelected(true);
        }
    }
    
    
    
   /**
    *    Sets up the log method radio buttons.
    */
    
    public void logMethod()
    {
        CBPanel logMethodPanel = new CBPanel();

        logMethodPanel.addln(new JLabel(CBIntText.get("Select a New Log Method: ")));
        logMethodPanel.addln(new JLabel(" "));    //TE: white space.

        logMethodCombo = new JComboBox(logMethodVal);
        logMethodCombo.setToolTipText(CBIntText.get("Set the log method in JXplorer."));

        logMethodPanel.addln(logMethodCombo);
        logMethodPanel.addln(new JLabel(" "));    //TE: white space.

        getLogMethod();

        tabbedPane.addTab(CBIntText.get("Log Method"), new ImageIcon(dirImage + "log_method.gif"), logMethodPanel, CBIntText.get("Set the method of logging you want, for example, to a file."));
    }



   /**
    *    Gets the log method from the property file and sets the appropriate radio button.
    */
    
    public void getLogMethod()
    {
        // remember that the config file value is stored in English...
        for(int i=0; i<logMethodValEnglish.length; i++)
        {
            if (logMethodValEnglish[i].equalsIgnoreCase(JXplorer.getProperty("logging")))
                logMethodCombo.setSelectedItem(logMethodVal[i]);
        }
    }    
    
    
    
   /**
    *    Sets up the log level radio buttons.
    */
    
    public void logLevel()
    {
        CBPanel logLevelPanel = new CBPanel();

        logLevelPanel.addln(new JLabel(CBIntText.get("Select a New Log Level: ")));
        logLevelPanel.addln(new JLabel(" "));

        logLevelCombo = new JComboBox(logLevelVal);
        logLevelCombo.setToolTipText(CBIntText.get("Set the logging level in JXplorer."));
        logLevelPanel.addln(logLevelCombo);
        logLevelPanel.addln(new JLabel(" "));

        getLogLevel();

        tabbedPane.addTab(CBIntText.get("Log Level"), new ImageIcon(dirImage + "log_level.gif"), logLevelPanel, CBIntText.get("Set the level of logging you want, for example, errors only."));
    }



   /**
    *    Gets the log level from the property file and sets the appropriate combo box item.
    */
    
    public void getLogLevel()
    {
        String logLevels = JXplorer.getProperty("debuglevel");    //TE: get the log level from the property file.

        switch (Integer.parseInt(logLevels))
        {
            case 0: { logLevelCombo.setSelectedItem(logLevelVal[0]); break; }       //TE: Errors Only option.
            case 1: { logLevelCombo.setSelectedItem(logLevelVal[1]); break; }       //TE: Basic option.
            case 2:
            case 3:
            case 4: { logLevelCombo.setSelectedItem(logLevelVal[2]); break; }       //TE: Tree Operations option.
            case 5:
            case 6: { logLevelCombo.setSelectedItem(logLevelVal[3]); break; }       //TE: Extensive option.
            case 7:
            case 8: { logLevelCombo.setSelectedItem(logLevelVal[4]); break; }       //TE: All option.
            case 9: { logLevelCombo.setSelectedItem(logLevelVal[5]); break; }       //TE: All + BER Trace option.
            default: { logLevelCombo.setSelectedItem(logLevelVal[0]); break; }      //TE: Errors Only option.
        }
    }



   /**
    *    Sets up the LDAP Levels tab.  Adds the text fields & labels to the panel.
    */
    
    public void ldapLevels()
    {
        ldapLimit   = new JTextField();
        ldapTimeout = new JTextField();        
        
        getLdapLevels();         
        
        CBPanel ldapLevelsPanel = new CBPanel();
        
        ldapLevelsPanel.addln(new JLabel(CBIntText.get("Set LDAP Options:           ")));
        ldapLevelsPanel.addln(new JLabel(" "));        //TE: white space.        
        ldapLevelsPanel.add(new JLabel(CBIntText.get("LDAP Limit: ")));
        ldapLevelsPanel.add(ldapLimit);
        ldapLevelsPanel.newLine();
        ldapLevelsPanel.addln(new JLabel(" "));        //TE: white space.
        ldapLevelsPanel.add(new JLabel(CBIntText.get("LDAP Timeout: ")));
        ldapLevelsPanel.add(ldapTimeout);        
                
        tabbedPane.addTab(CBIntText.get("Search Limits"), new ImageIcon(dirImage+"find.gif"), ldapLevelsPanel, CBIntText.get("Set the search levels, that is, the number of entries returned from a search and the timeout."));
    }
       
     
    
   /**
    *    Gets the ldap limit and timeout values from the property file and sets the text fields with these values.
    */
    
    public void getLdapLevels()
    {
        String limit = JXplorer.getProperty("option.ldap.limit");        //TE: gets the values from the property file.
        String timeout = JXplorer.getProperty("option.ldap.timeout");    

        ldapLimit.setText(limit);
        ldapLimit.setToolTipText(CBIntText.get("Enter the new limit level."));
        ldapTimeout.setText(timeout);
        ldapTimeout.setToolTipText(CBIntText.get("Enter the new timeout level."));
    }



   /**
    *    Sets up the URL tab with a combo box.
    */

    public void urlTab()
    {
        String[] url = new String[] {CBIntText.get("JXplorer"),CBIntText.get("Launch")};

        CBPanel urlPanel = new CBPanel();
        urlCombo = new JComboBox(url);

        getURLHandling();

        urlPanel.addln(new JLabel(CBIntText.get("Select URL handling: ")));
        urlPanel.addln(new JLabel(" "));    //TE: white space.
        urlPanel.addln(urlCombo);
        urlPanel.addln(new JLabel(" "));    //TE: white space.
        urlPanel.addln(new JLabel(CBIntText.get("Note: Launch is for Windows only.")));

        tabbedPane.addTab(CBIntText.get("URL"), new ImageIcon(dirImage + "url.gif"), urlPanel, CBIntText.get("Select how you would like the URLs handled in JXplorer."));
    }



   /**
    *   Gets the url handling type (either JXplorer or Launch - if something else, it defaults to JXplorer)
    *   from the property file and sets the url combo box.
    */

    public void getURLHandling()
    {
        String urlHandling = JXplorer.getProperty("option.url.handling");        //TE: gets the value from the property file.

        int index = 0;

        if(urlHandling!=null)
            if(urlHandling.equalsIgnoreCase("Launch"))
                index = 1;

        urlCombo.setSelectedIndex(index);
    }



   /**
    *    Calls the appropriate save methods then exits.
    */

    public void apply()
    {
        checkLookAndFeel();
        checkLogMethod();
        checkLogLevel();
        checkLdapLevels();
        checkUrlHandling();

        quit();
    }
    
    
    
   /**
    *    Checks if the L&F has been changed.  If so updates the property file and sets the L&F as requested.
    */
    
    public void checkLookAndFeel()
    {        
        getOwner().setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));    //TE: sets the cursor to an hour-glass.
        JRadioButton rb = null;    //TE: used for exception info.
                
        try
        {
           if ((lookAndFeel[WINDOWS].isSelected()) == true)
           {
               rb = lookAndFeel[WINDOWS];        //TE: if this fails we know which L&F made it fail.
               UIManager.setLookAndFeel("com.sun.java.swing.plaf.windows.WindowsLookAndFeel"); 
               JXplorer.setProperty("gui.lookandfeel", "com.sun.java.swing.plaf.windows.WindowsLookAndFeel");    //TE: updates the 'gui.lookandfeel' value in the dxconfig property file with the windows L&F.
           }  
           else if((lookAndFeel[JAVA].isSelected()) == true)
           {
               rb =lookAndFeel[JAVA];            //TE: if this fails we know which L&F made it fail.
               UIManager.setLookAndFeel("javax.swing.plaf.metal.MetalLookAndFeel"); 
               JXplorer.setProperty("gui.lookandfeel", "javax.swing.plaf.metal.MetalLookAndFeel");    //TE: updates the 'gui.lookandfeel' value in the dxconfig property file with the java L&F.
           }
           else if((lookAndFeel[MOTIF].isSelected()) == true)
           {
               rb = lookAndFeel[MOTIF];          //TE: if this fails we know which L&F made it fail.
               UIManager.setLookAndFeel("com.sun.java.swing.plaf.motif.MotifLookAndFeel"); 
               JXplorer.setProperty("gui.lookandfeel", "com.sun.java.swing.plaf.motif.MotifLookAndFeel");    //TE: updates the 'gui.lookandfeel' value in the dxconfig property file with the motif L&F.    
           }
           else if((lookAndFeel[MAC].isSelected()) == true)
           {
               rb = lookAndFeel[MAC];          //TE: if this fails we know which L&F made it fail.
               UIManager.setLookAndFeel("com.sun.java.swing.plaf.mac.MacLookAndFeel");
               JXplorer.setProperty("gui.lookandfeel", "com.sun.java.swing.plaf.mac.MacLookAndFeel");    //TE: updates the 'gui.lookandfeel' value in the dxconfig property file with the motif L&F.
           }

            updateLookAndFeel();

            if (mainMenu.getConnection != null) SwingUtilities.updateComponentTreeUI(mainMenu.getConnection);
        }
        catch (UnsupportedLookAndFeelException exc)  // it can throw an exception if you try to set windows on a non-windows box.
        {
            rb.setEnabled(false);
            CBUtility.log("Unsupported LookAndFeel: " + rb.getText());
        }
        catch (Exception exc)  // shouldn't happen, but this is a pretty bizarre operation so just in case...
        {
            rb.setEnabled(false);
            exc.printStackTrace();
            CBUtility.log("Could not load LookAndFeel: " + rb.getText());
            exc.printStackTrace();
        }
        getOwner().setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));    //TE: sets the cursor back to normal.
		
		(jx.getAttributeDisplay()).refreshEditors();	//TE: force a refresh of the attribute display...so that the fields don't disappear (bug 367).		
    }



   /**
    *   There must be a better way of doing this but here it is anyway...
    *   This method tries to make sure that the whole of JXplorer actually gets
    *   the new Look and Feel.  It should work by just using...
    *   <br>'SwingUtilities.updateComponentTreeUI(getOwner());',<br>
    *   but some components are forgotten about - such as this window, the search
    *   gui and the popup tree tool.  So this method manually tells them to update.
    *   But thats not all - each tree has its own copy of the search gui & the popup
    *   tree tool...so each tree needs to be told to update these components.
    *   In all there are eight calls to the updateComponentTreeUI...phew.
    */

    private void updateLookAndFeel()
    {
        if((getOwner() instanceof JXplorer) == false)
        {
            SwingUtilities.updateComponentTreeUI(getOwner());
            return;
        }

        JXplorer jx = (JXplorer)getOwner();

//TE XXXXXXXXXX this produces bug 2578..........
        //TE: update the bulk of JXplorer with the new L&F...
        SwingUtilities.updateComponentTreeUI(jx);

        //TE: make sure this window gets the new L&F too...
        SwingUtilities.updateComponentTreeUI(this);

        //TE: get the trees...
        SmartTree explore = jx.getTree();
        SmartTree search = jx.getSearchTree();
        SmartTree schema = jx.getSchemaTree();

        //TE: to stop the truncation of tree nodes esp in Java L&F...what a joke....
        SwingUtilities.updateComponentTreeUI(explore);
        SwingUtilities.updateComponentTreeUI(search);
        SwingUtilities.updateComponentTreeUI(schema);

        //TE: make sure each tree's popup menu gets the new L&F...
        SwingUtilities.updateComponentTreeUI(explore.getPopupTool());
        SwingUtilities.updateComponentTreeUI(search.getPopupTool());
        SwingUtilities.updateComponentTreeUI(schema.getPopupTool());

        //TE: make sure each tree's search dialog gets the new L&F...
        SearchGUI sExplore = explore.search;
        SearchGUI sSearch = search.search;
        SearchGUI sSchema = schema.search;

        if(sExplore!=null)
            SwingUtilities.updateComponentTreeUI(sExplore);
        if(sSearch!=null)
            SwingUtilities.updateComponentTreeUI(sSearch);
        if(sSchema!=null)
            SwingUtilities.updateComponentTreeUI(sSchema);
    }



    /**
    *    Checks if the log method has been changed.  If so updates the property file and sets the log method as requested.
    */
    
    public void checkLogMethod()
    {
        try
        {
            String method = logMethodValEnglish[logMethodCombo.getSelectedIndex()];
            JXplorer.setProperty("logging", method);                      //TE: sets the value of the selected radio button in the property file.
            jx.checkPropertyDependantActions("logging", method);    //TE: sets the values for immediate use.
        }
        catch (Exception e)
        {
            return;  // should never happen ... ho ho.  TODO: log properly
        }

    }

  
    
   /**
    *    Checks if the log level has been changed.  If so updates the property file and sets the log level as requested.
    */
    
    public void checkLogLevel()
    {
        int level = 0;

        switch (logLevelCombo.getSelectedIndex())
        {
            case 0: { level=0; break; }       //TE: Errors Only option.
            case 1: { level=1; break; }       //TE: Basic option.
            case 2: { level=4; break; }       //TE: Tree Operations option.
            case 3: { level=6; break; }       //TE: Extensive option.
            case 4: { level=8; break; }       //TE: All option.
            case 5: { level=9; break; }       //TE: All + BER Trace option.
            default: { level=0; break; }      //TE: Errors Only option.
        }

        JXplorer.setProperty("debuglevel", String.valueOf(level));                      //TE: sets the value of the selected radio button in the property file.
        jx.checkPropertyDependantActions("debuglevel", String.valueOf(level));          //TE: sets the values for immediate use.
    }
    
    
    
   /**
    *    Checks if the ldap levels have been changed by getting the values from the text areas.
    *    Sets the changes in the property file and in searchBroker.
    */
    
    public void checkLdapLevels()
    {
        String limit = ldapLimit.getText();        
        String timeout = ldapTimeout.getText();
        
        try            //TE: make sure the values are of integer type.
        {
            Integer.valueOf(limit);
            Integer.valueOf(timeout);
        }
        catch(NumberFormatException e)
        {
            CBUtility.error("Both Ldap Limit & Ldap timeout must be of Integer type.\n" + e);
            getLdapLevels();
        }
       
        JXplorer.setProperty("option.ldap.limit", limit);               //TE: set the values in the property file.
        JXplorer.setProperty("option.ldap.timeout", timeout);

        jx.searchBroker.setTimeout(Integer.parseInt(timeout));    //TE: sets the values in searchBroker for immediate use.
        jx.searchBroker.setLimit(Integer.parseInt(limit));          
    }



   /**
    *   Checks what the URL handling is and sets it in the property file.  Does this by getting the
    *   user selected index...if 0 then 'JXplorer' if 1 then 'Launch'.
    */

    public void checkUrlHandling()
    {
        int index = urlCombo.getSelectedIndex();

        if(index == 1)
            JXplorer.setProperty("option.url.handling", "Launch");
        else
            JXplorer.setProperty("option.url.handling", "JXplorer");
    }



   /**
    *    Resets all values in the advanced options window back to the values in the property file.
    */
    
    public void reset()
    {
        getLookAndFeel();
        getLogLevel();
        getLogMethod();            
        getLdapLevels();
        getURLHandling();
    }
    
    
      
   /**
    *    Shuts the advanced options dialog.
    */
        
    public void quit()
    {
        setVisible(false);
        dispose();
    }
}
