package com.ca.directory.jxplorer;

import javax.swing.*;

import com.ca.directory.jxplorer.broker.*;
import com.ca.commons.cbutil.*;
import com.ca.commons.jndi.ConnectionData;

import org.apache.log4j.*;

import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.net.URISyntaxException;


/**
*   JXOpenConWin allows the user to open an ldap connection.  The jndiBroker
*   that the connection is opened through is attached to a tree display, and
*   a separate schema tree display.  An attribute viewer is linked with the
*   main tree display.  A JLabel allows the connection window to communicate
*   status to the user.  These parameters are all obtained directly from the
*   main JXplorer object
*	@author Trudi
*/

public class JXOpenConWin extends CBOpenConWin implements DataListener
{
    private final static Logger log = Logger.getLogger("com.ca.directory.jxplorer.CBOpenConWin");

	private JXplorer jxplorer;
	private JNDIBroker jndiBroker;
	private String referral;
	private String aliasType;
    private JTextField dsmlService;

    /**
     *  Constant used to add 'DSML' option to combo box.
     */
    protected static final String DSMLV2 = CBIntText.get("DSML v2");


   /**
    *	<p>JXOpenConWin allows the user to open an ldap connection.  The jndiBroker
	*  	that the connection is opened through is attached to a tree display, and
	*  	a separate schema tree display.  An attribute viewer is linked with the
	*   main tree display.  A JLabel allows the connection window to communicate
	*   status to the user.  These parameters are all obtained directly from the
	*   main JXplorer object</p>
	*   <p>
	*   @param jx a JXplorer object to obtain trees, data brokers and gui hooks from.
	*	@param statusDisplay a label used to give status feedback to the user.
	*   @param clientcerts the client certificate keystore (optional if 'simple ssl' is used).
	*   @param cacerts the trusted server certificate keystore (required for ssl)
	*   @param referral the jndi referral handling method ("follow" is default).
	*   @param aliasType the jndi alias handling - whether aliases are searched or not.
	*           (default is "searching");
	*/



    public JXOpenConWin(JXplorer jx, JLabel statusDisplay, String clientcerts, String cacerts,
							String referral, String aliasType)
    {
		super(jx, statusDisplay, clientcerts, cacerts, referral, aliasType);

		jxplorer = jx;
        newCon.tracing = jxplorer.jndiBroker.getTracing();
    }


    /**
     *
     * @param statusDisplay
     */

    protected void initGUI(JLabel statusDisplay)
    {

        super.initGUI(statusDisplay);

        // make 'dsml available' the default, but allow it to be turned off for light weight distributions...
        if (JXplorer.getProperty("dsml") == null ||
                JXplorer.getProperty("dsml").equalsIgnoreCase("false")==false)
            version.addItem(DSMLV2);

        myTemplater.loadDefault();  // needs to be redone after 'dsml' added to property list

        display.validate();

    }



    /**
     * This method overrides the CBOpenConWin.addExtraComponent method to insert
     * the DSML URL label and text field.
     * .
     */

    public void addExtraComponent()
    {
        JLabel urlLabel;
        display.makeLight();
        display.add(urlLabel = new JLabel("   " + CBIntText.get("DSML Service") + ": "), 0,2,1,1);
        display.addWide(dsmlService = new JTextField("", 30),4);
        urlLabel.setToolTipText(CBIntText.get("The DSML service; e.g. ") + "'axis/dsml' or 'dsml-sample/services/DSML'");

        VersionActionListener versionListener = new VersionActionListener();
        version.addActionListener(versionListener);
    }



   /**
    * Implements ActionListener to enable or disable the
    * DSML text field depending on protocol. If either LDAP V2 or V3
    * is selected...the text field will be disabled.  And the security levels -
    * if DSML is selected the security combo is disabled.
	* @author Trudi.
	*/

	class VersionActionListener implements ActionListener
	{

	   /**
	    *	Enables/disables the DSML text field depending on protocol.
        *   And the security levels.
		*/

		public void actionPerformed(ActionEvent event)
		{
            if(!version.getSelectedItem().equals(DSMLV2))
            {
                dsmlService.setEnabled(false);
                dsmlService.setText("");
                dsmlService.setBackground(Color.lightGray);

                level.setEnabled(true);
                checkSecurityLevel();
            }
            else
            {
                dsmlService.setEnabled(true);
                dsmlService.setBackground(Color.white);

                level.setSelectedIndex(0);
                managerDN.setText("");
                password.setText("");
                checkSecurityLevel();
                level.setEnabled(false);

                /*
                managerDN.setEnabled(false);
                managerDN.setBackground(Color.lightGray);
                password.setEnabled(false);
                password.setBackground(Color.lightGray);
                */
            }
		}
	}



    /**
     *   Over-ride base 'doOK()' method to allow for setting the new 'DSML' protocol if selected...
     */

    public void doOK()
    {
        if (version.getSelectedItem().equals(DSMLV2))
            newCon.protocol = ConnectionData.DSML;
        else
            newCon.protocol = ConnectionData.LDAP;

        super.doOK();

    }



   /**
    *  	This resets some volitile parameters that might have changed
    *   since the last time the user accessed the window.
    *
    *   @param newclientcerts the client certificate keystore (optional if 'simple ssl' is used).
    *   @param newcacerts the trusted server certificate keystore (required for ssl)
    *   @param newreferral the jndi referral handling method ("follow" is default).
    *   @param newaliasType the jndi alias handling - whether aliases are searched or not.
    *           (default is "searching");
    */

    public void reinit(String newclientcerts, String newcacerts,
                            String newreferral, String newaliasType)
    {
        newCon.clientcerts 	= newclientcerts;
        newCon.cacerts 		= newcacerts;
        newCon.referralType = newreferral;
        newCon.aliasType 	= newaliasType;

        if(jndiBroker==null)    //TE: bug 3222.
            newCon.tracing  = jxplorer.jndiBroker.getTracing();
        else
            newCon.tracing  = jndiBroker.getTracing();
    }



    /**
     * Set's title back to 'open ldap connection'/clears password...
     * but sets the title to 'Open LDAP/DSML Connection".
     */

    public void resetTitleAndPassword()
    {
        this.setTitle(CBIntText.get("Open LDAP/DSML Connection"));
        password.setText("");
    }



   /**
	*   This method is called when the user connection data
	*   has been gathered and (roughly) checked.
	*	@param connectData the parsed connection data containing
	*          host and port details, security info, etc.
	*/

	public void connect(ConnectionData connectData)
	{
        // Queue the request on the broker's request list.  When completed, the request
        // will trigger the 'DataReady()' method below.

        try
        {
            jndiBroker = jxplorer.jndiBroker;

            // clear the GUI preparatory to the new data appearing.
            jxplorer.preConnectionSetup();

            // create a new data query and put it on the query stack for the connection thread.
            DataQuery query = jndiBroker.connect(connectData);

            query.addDataListener(this);

        }
        // the code above just sets up the connection call, and doesn't really
        // do anything that should cause an exception - the connection is
        // done in a different thread, and any exceptions should be caught
        // by the 'DataReady()' code below.

        catch (Exception e)
        {
            CBUtility.log("Unexpected exception in JXOpenConWin.connect: " + e, 0);
            e.printStackTrace();
        }
	}

   /**
	*    This method is called when a new connection attempt is finished.
	*    If successfull, the main, search and schema trees are updated
	*    with theire data sources, and a successfull connection message
	*    displayed to the user.  If unsuccessful, an error message is displayed.
	*    @param request the connection request data object.
	*/

    //XXX remember that this method is being run by the directory connection thread -
    //XXX so it can use unthreaded jndiBroker directory methods with impunity.

    public void dataReady(DataQuery request)
    {
        if (!(request instanceof JNDIBroker.DataConnectionQuery))
		{
			log.warn("Incorrect data for connection - cannot connect");
			return;
		}

        if (request.hasException()==false)  // apparently we have a valid connection to play with!
        {
            if (jxplorer.postConnectionSetup((JNDIBroker.DataConnectionQuery)request))
            {
                setVisible(false);

				((JNDIBroker.DataConnectionQuery)request).conData.clearPasswords();

                dispose();
            }
            else
            {
                jxplorer.disconnect();
            }
        }
        else    // request registered an error in jdniBroker
        {
            // do we want a nice error box?  I think we do!
            new CBErrorWin(this, "Error opening connection:\n" + request.getException().getMessage(), request.getException()); // automatically visible one-shot.

            log.error("Error opening connection ", request.getException());
            request.clearException();
            setTitle(CBIntText.get("Couldn't Connect" + ": " + "Try Again"));
			dispose();	//TE: don't remove this...for some reason if an incorrect port etc is entered on Linux - JX crashes!
			show();
            userMessage.setText(CBIntText.get("Couldn't Open ") + request.getExtendedData("url"));
            jxplorer.disconnect();
            request.squelch();  // we're done here.
        }
    }

    /**
     * overload the getURL method of the base class to use the new url field.
     * @return the correct, parsed URL.
     */
    protected String getURL() throws NumberFormatException, URISyntaxException
    {
        String url = super.getURL();
        if (version.getSelectedItem().equals(DSMLV2)) // dsml may require an extra 'service' bit to add to the url...
        {
            String dsml = dsmlService.getText();
            if (dsml.startsWith("/"))               // trim any starting slash - we add it back just below.
                dsml = dsml.substring(1);

            if (url.startsWith("ldap://"))          // trim the ldap: protocol prefix - this is usually set by CBOpenConWin...
                url = url.substring(7);

            url = "http://" + url + "/" + dsml;     // construct the full dsml url...
        }
        log.debug("connecting with url: " + url);
        return url;
    }
}